#ifndef D3_H
#define D3_H
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "graphics.h"
#include "error.h"

extern "C" {
	#include "libsx.h"
}

typedef long fixed;                     // Our new fixed point type.

#define itofx(x) ((x) << 8)                   // Integer to fixed point
#define ftofx(x) ((long)((x) * 256))          // Float to fixed point
#define dtofx(x) ((x) * 256)                  // Double to fixed point
#define fxtoi(x) ((x) >> 8)                   // Fixed point to integer
#define fxtof(x) ((float) (x) / 256)          // Fixed point to float
#define fxtod(x) ((double)(x) / 256)          // Fixed point to double
#define Mulfx(x,y) (((y) * (x)) >> 8)         // Multiply a fixed by a fixed
#define Divfx(x,y) ((x << 8) / (y))           // Divide a fixedby a fixed
#define Printfx(x) printf("%ld.%ld\n", x >> 16, 10000 * (unsigned long) ((x) & 0xffff) >> 16)
// Print fixed point.
#define NDPrintfx(x) printf("%ld", x >> 16)
// Print fixed point without a decimal point.


/* #define PRECALC_NORMALS */

#define VIEWPORT_MIN_X -100.0
#define VIEWPORT_MIN_Y -100.0
#define VIEWPORT_MIN_Z -100.0

#define VIEWPORT_MAX_X 100.0
#define VIEWPORT_MAX_Y 100.0
#define VIEWPORT_MAX_Z 100.0

#define FindZ( normal, xx, yy, vertex ) (float)(xx*normal->x + yy*normal->y + ((-1.0*normal->x)*vertex->x-normal->y*vertex->y - normal->z*vertex->z))/(float)normal->z;

void InitZBuffer( void );

/*** file parsing stuff ***/
char *ReturnFirstPart( char *str );
char *ReturnLastPart( char *str );
char *GetWord( char *str, int *pos );
char **Parse( char *str );
/**************************/

typedef struct matrix_tag{
        float matrix[4][4];
	int rows;
	int cols;
} matrix_type;

matrix_type *MatrixMultiply4x4_X_4x1( matrix_type *a, matrix_type *b );
matrix_type *MatrixMultiply4x4_X_4x4( matrix_type *a, matrix_type *b );
                       
typedef struct Vertex_tag{
        long id;

        float x;
        float y;
        float z;
        float w;

        float orig_x;
        float orig_y;
        float orig_z;
        float orig_w;

        struct Vertex_tag *normal;
	struct Vertex_tag *texture_normal;

        struct Vertex_tag *next;
} Vertex;


typedef struct camera_tag{
	Vertex *position;
	Vertex *focus;
	Vertex *u, *v, *n;
} Camera;


void Init3DEngine( int, int );


Camera *MakeCamera( void );


Vertex *NewVertex ( float x, float y, float z );
Vertex *NewVertex ( float x, float y, float z, float w );
Vertex *NewVertex( float x, float y, float z, float xn, float yn, float zn );
void   DestroyList( Vertex **vertex );

void Normalize( Vertex *vector );
                              
/***** begin: used by ScanFill ********/

void   SortScan   ( Vertex *list );
float FindHighY  ( Vertex *vertex );
float FindLowY   ( Vertex *vertex );
Vertex *GetIntersection( Vertex *line1point1, Vertex *line1point2, Vertex *line2point1, Vertex *line2point2 );
float DotProduct( Vertex *vector1, Vertex *vector2 );
void ScanFill( Vertex *vertex, Vertex *normal, int low_color, int high_color, Image *where );

/******** end *********************************/


class Face{
        public:
        Vertex *begin;
        Vertex *end;

        Vertex *begin_per;      /* seperate list for perspective */
        Vertex *end_per;

        float use_perspective;
        char remove_backfaces;
	char use_texture;
	char use_stretched_texture;
	char in_front_of_camera;

       // public:

        int low_color;
        int high_color;
        Vertex *vertex;
        Vertex *in_perspective;
        Vertex *normal_to_face;

	Image texture;
	paletteType palette;

        Face( void );
        ~Face( void );

        char MoveToVertex( int id );
        char MoveToNormal( int id );

        void MoveFirst( void );
        char MoveNext( void );
        void MoveLast( void );

        void MoveFirstPer( void );
        char MoveNextPer( void );
        void MoveLastPer( void );
        
        void AddVertex( float x, float y, float z );
        void AddVertex( float x, float y, float z, float xn, float yn, float zn );
        void Remove( void );
        char IsEmpty( void );

        void DrawSolidNoShade( Image *where );
        void DrawSolidFlatShade( Image *where );
        void DrawSolidGouraudShade( Image *where );
        void DrawSolidPhongShade( Image *where );
        void DrawSolidFastPhongShade( Image *where, char *phongmap );
        void DrawSolidTransparentShade( Image *where );
        void DrawSolidTexture( Image *where );
        void DrawSolidStretchedTexture( Image *where );
        void DrawWireFrameNoPerspective( Image *where );
        void DrawWireFrameInPerspective( Image *where );
        void RotateX( float angle );
        void RotateY( float angle );
        void RotateZ( float angle );
        void AddPerspective( float d );
        void Move( float dx, float dy, float dz );
	void Scale( float amount );
        Vertex *FindNormal( void );
        Vertex *FindNormalPer( void );
        char Visible( void );

	void AddCamera( matrix_type *m );

        void UsePerspective( float value );
        void RemoveBackfaces( char value );

	void UseTexture( char *filename );
	void UseStretchedTexture( char *filename );

	void ChangeColor( int low, int high );
};

typedef struct Face_list_tag {
        Face *face;

        struct Face_list_tag *next;
} Face_list;

class Object_3d{
        public:
        int low_color;
        int high_color;
        Face_list *begin;
        Face_list *end;

        float use_perspective;
        char remove_backfaces;
        char use_no_shade;
        char use_flat_shade;
        char use_gouraud_shade;
        char use_phong_shade;
	char use_fastphong_shade;
	char use_transparent_shade;
	char use_texture;
	char use_stretched_texture;

	char *phongmap;

        //public:

        Face_list *face;

        Object_3d( char *filename );
        Object_3d( void );
        ~Object_3d( void );


        void MoveFirst( void );
        char MoveNext( void );
        void MoveLast( void );

        void AddFace( Face *add );
        void Remove( void );
        char IsEmpty( void );

        void DrawWireFrame( Image *where );
        void DrawSolid( Image *where );
        void DrawWireFrameNoPerspective( Image *where );
        void DrawWireFrameInPerspective( Image *where );
        void RotateX( float angle );
        void RotateY( float angle );
        void RotateZ( float angle );
        void AddPerspective( float d );
        void Move( float dx, float dy, float dz );
        void CalculateNormals( void );
        void CalculateNormalsForVertices( void );
	void GroupVertexIDs( void );
	void ShowGroupVertexIDs( void );

        void UsePerspective( float value );
        void RemoveBackfaces( char value );
        void UseNoShade( void );
        void UseFlatShade( void );
        void UseGouraudShade( void );
        void UsePhongShade( void );
	void UseFastPhongShade( void );
	void UseTransparentShade( void );
	void UseTexture( char *filename );
	void UseTexture( Image *image );
	void UseStretchedTexture( Image *image );
	void UseStretchedTexture( char *filename );

	void SetCurrentAsOrig( void );
	void Reset( void );

	void AddCamera( Camera *camera );

	void Scale( float amount );

	void ChangeColor( int low, int high );

	void CalcPhongMap( void );

	void Tween( Object_3d *object, float t );
};
#endif
